@@ -4,20 +4,6 @@
 //
 "use strict";
 
-/**
- * Redirect to the studio to create a new task
- */
-function studio_create_new_task()
-{
-    var task_id = $('#new_task_id');
-    if(!task_id.val().match(/^[a-zA-Z0-9_\-]+$/))
-    {
-        alert('Task id should only contain alphanumeric characters (in addition to "_" and "-").');
-        return;
-    }
-    window.location.href = window.location.href + "/../edit/task/" + task_id.val()
-}
-
 /**
  * Load the studio, creating blocks for existing subproblems
  */
@@ -45,7 +31,7 @@ function studio_load(data)
     });
 
     // Must be done *after* the event definition
-    if(collapsable.length != 1)
+    if(collapsable.length !== 1)
         collapsable.collapse('hide');
 
     $('form#edit_task_form').on('submit', function()
@@ -53,6 +39,15 @@ function studio_load(data)
         studio_submit();
         return false;
     });
+
+    studio_update_environments();
+    $('#environment-type').change(studio_update_environments);
+}
+
+function studio_update_environments() {
+    var env_type = $('#environment-type').val();
+    $('.environment-boxes').hide();
+    $('#environment-box-'+env_type).show();
 }
 
 /**
@@ -750,4 +745,36 @@ function uploadData(formdata){
             console.log("something went wrong");
         }
     });
+}
+
+// Use selectize.js to select for a user.
+// element is the element to be used as a selector ($('#id'))
+// path_to_search_user is the path to the search_user page (/admin/tutorial/search_user/) with the final /
+// current_users is an array of dict with the format [{'username': 'theusername', 'realname': 'The realname'}]
+// single: true for single user selection, false for multiple.
+function user_selection(element, path_to_search_user, current_users, single) {
+    element.selectize({
+        delimiter: ',',
+        persist: false,
+        valueField: 'username',
+        labelField: 'realname',
+        searchField: ['username', 'realname'],
+        create: false,
+        options: current_users.map(function(x) { return {"username": x.username, "realname": x.realname + " (" + x.username + ")"} }),
+        items: current_users.map(function(x) { return x.username }),
+        load: function(query, callback) {
+            if (!query.length) return callback();
+            $.ajax({
+                url: path_to_search_user + encodeURIComponent(query),
+                type: 'GET',
+                error: function() {
+                    callback();
+                },
+                success: function(res) {
+                    callback(res[0].map(function(x) { return {"username": x.username, "realname": x.realname + " (" + x.username + ")"} }));
+                }
+            });
+        },
+        maxItems: single ? 1 : null
+    });
 }
\ No newline at end of file
